/**
 * Model class that maps to elasticsearch document scoped by index and type
 * @class ElasticSearch
 */
class ElasticSearch {
    /**
     * Default constructor
     * @constructor
     */
    constructor() {
        /**
         * primary id for the document saved in elasticsearch
         * @type {id|string|null}
         * @private
         */
        this._primary_id = null;
        /**
         * Query statement used to search against elasticsearch
         * @type {*}
         * @private
         */
        this._query = {match_all : {}};

        /**
         * Number of records to skip
         * @type {number}
         * @private
         */
        this._offset = 0;
        /**
         * Maximum number of records to pull from elasticsearch
         * @type {number}
         * @private
         */
        this._limit = 50;
        /**
         * Filters to filter records. This is used with query together.
         * @type {}
         * @private
         */
        this._filters = {};

        /**
         * Fields to sort the order. By default, the records are sorted by _score in descending order
         * @type {{_score: {order: string}}[]}
         * @private
         */
        this._sort = [{'_score' : {'order':'desc'}}];

    }

    static find() {
        return new this();
    }

    /**
     * Delete all records by a given query and filter
     * @method deleteAll
     * @async
     * @return {Promise<void>}
     */
    deleteAll(refresh = false) {
        let query = [];
        if(Object.keys(this._filters)>0) {
            query = {
                bool : {
                    filter: this._filters,
                    must: this._query,
                }
            };
        } else {
            query = this._query;
        }
        let params = {
            index: this.constructor.index,
            type: this.constructor.type,
            body: {
                query: query,
            },
            refresh : refresh
        };
        return this.constructor.client.deleteByQuery(params);
    }

    range(field, conditions, should_or_must = "must") {
        if(Object.keys(this._query).length === 0) {
            this._query = {
                range : {}
            };
            this._query.range[field] = conditions;
        } else {
            let old_queries = this._query;
            let query = {};
            query[field] = conditions;
            let new_query = {
                range : query
            };
            if(should_or_must === 'must') {
                this._query = {
                    bool : {
                        must : []
                    }
                };
            } else {
                this._query = {
                    bool : {
                        should : []
                    }
                };
            }
            this._query['bool'][should_or_must].push(old_queries);
            this._query['bool'][should_or_must].push(new_query);
        }
        return this;
    }

    /**
     * Delete a record by a given id
     * @method delete
     * @async
     * @param id
     * @return {Promise<*>}
     */
    static delete(id, refresh = false) {
        let params = {
            index: this.index,
            type: this.type,
            id: id,
            refresh : refresh
        };
        return this.client.delete(params);
    }

    /**
     * Update mapping
     * @method updateMapping
     * @async
     * @return {Promise<void>}
     */
    static updateMapping() {
        return this.client.indices.putMapping(this.mapping);
    }

    static indexExists(local = false) {
        let params = {
            index : this.index,
            local : local
        };
        return this.client.indices.exists(params);
    }

    static exists(id, refresh = false) {
        let params = {
            index: this.index,
            type: this.type,
            id: id,
            refresh: refresh
        };
        return this.client.exists(params);
    }

    /**
     * save a document
     * @method save
     * @async
     * @return {Promise<any>}
     */
    async save(refresh = false) {
        let index_exists = await this.constructor.indexExists();
        if(!index_exists) {
            await this.constructor.createIndex();
        }
        let exists = await this.constructor.exists(this._primary_id, refresh);
        if(!exists) {
            return await add.call(this, refresh);
        } else {
            return await update.call(this, refresh);
        }
    }

    /**
     * return attributes of a model
     * @method attributes
     * @return {string[]}
     */
    static attributes()
    {
        let type = this.type;
        let mapping = this.mapping;
        return Object.keys(mapping[type]['properties']);
    }


    /**
     * Create index
     * @method createIndex
     * @async
     * @return {Promise<void>}
     */
    static createIndex() {
        let params = {
            index: this.index,
            'body': {
                'mappings': this.mapping
            }
        };
        return this.client.indices.create(params);
    }

    /**
     * Delete a model's index
     * @method deleteIndex
     * @return {Promise<*>}
     */
    static deleteIndex() {
        return this.client.indices.delete({index:this.index});
    }

    /**
     * Build filters
     * @param conditions
     * @param should_or_must could be 'should' or 'must'
     * @return {ElasticSearch}
     */
    filter(conditions, should_or_must = "must") {
        this._filters = {};
        let filters = [];
        for(let key in conditions) {
            let value = conditions[key];
            if(Array.isArray(value)) {
                if(value.length === 1) {
                    let term = {};
                    term[key] = value[0];
                    filters.push({'term' : term});
                } else if(value.length === 0) {
                    return this;
                } else {
                    let filter = {
                        'bool': {
                            'should': []
                        },
                    };
                    value.forEach(function(element){
                        let term = {};
                        term[key] = element;
                        filter['bool']['should'].push({'term' : term});
                    });
                    filters.push(filter);
                }
            } else {
                let term = {};
                term[key] = value;
                filters.push({'term' : term});
            }
        }
        if(filters.length === 1) {
            this._filters = filters[0];
        } else {
            resetFilters.call(this, should_or_must);
            this._filters['bool'][should_or_must] = filters;
        }
        return this;
    }

    /**
     * Add boolean should filters
     * @method orFilter
     * @param conditions
     * @return {ElasticSearch}
     */
    orFilter(conditions, should_or_must = 'must') {
        if(Object.keys(this._filters).length===0) {
            return this.filter(conditions, should_or_must);
        }
        let old_filters = this._filters;
        this._filters={};
        this.filter(conditions);
        let new_filters = this._filters;
        let filter= {
            bool: {
                should: []
            },
        };
        filter['bool']['should'].push(old_filters);
        filter['bool']['should'].push(new_filters);
        this._filters = filter;
        return this;
    }

    /**
     * Add boolean must filters
     * @method orFilter
     * @param conditions
     * @return {ElasticSearch}
     */
    andFilter(conditions) {
        if(Object.keys(this._filters).length === 0) {
            return this.filter(conditions);
        }
        let old_filters = this._filters;
        this._filters={};
        this.filter(conditions);
        let new_filters = this._filters;
        let filter= {
            bool: {
                must: []
            },
        };
        filter['bool']['must'].push(old_filters);
        filter['bool']['must'].push(new_filters);
        this._filters = filter;
        return this;
    }

    /**
     * Add sort field to search query
     * @method orderBy
     * @param conditions
     * @return {ElasticSearch}
     */
    orderBy(conditions) {
        if(conditions.length>0) {
            this._sort = [];
            for(let key in conditions) {
                let value = conditions[key];
                this._sort.push({key : {order : value}});
            }
        } else {
            this._sort = [{_score : {order : 'desc'}}];
        }
        return this;
    }

    /**
     * add search query
     * @param conditions
     * @param should_or_must can be "must" or "should", default is "must"
     * @return {ElasticSearch}
     */
    query(conditions, should_or_must = 'must') {
        if(Array.isArray(conditions)) {
            if(conditions.length > 1) {

                if(should_or_must === 'must') {
                    this._query = {'bool': {must : []}};
                } else {
                    this._query = {'bool': {should : []}};
                }
                this._query['bool'][should_or_must] = conditions;
            } else {
                this._query = conditions[0];
            }
        } else if(typeof conditions === "object") {
            this._query = conditions;
        } else {
            this._query = {match_all : {}};
        }
        return this;
    }

    /**
     * Add boolean should query
     * @param conditions
     * @param should_or_must default is "must". Can be "must" or "should". Only needed when conditions is array
     * @returns {ElasticSearch}
     */
    orQuery(conditions, should_or_must = "must") {
        if(this._query.match_all!==undefined) {
            return this.query(conditions, should_or_must);
        }
        let old_queries = this._query;
        this._query={};
        this.query(conditions, should_or_must);
        let new_queries = this._query;
        let query= {
            bool: {
                should: []
            },
        };
        query['bool']['should'].push(old_queries);
        query['bool']['should'].push(new_queries);
        this._query = query;
        return this;
    }

    /**
     * Add boolean must query
     * @param conditions
     * @param should_or_must default is "must". Can be "must" or "should". Only needed when conditions is array
     * @returns {ElasticSearch}
     */
    andQuery(conditions, should_or_must = "must") {
        if(this._query.match_all !== undefined) {
            return this.query(conditions, should_or_must);
        }
        let old_queries = this._query;
        this._query={};
        this.query(conditions, should_or_must);
        let new_queries = this._query;
        let query= {
            bool: {
                must: []
            },
        };
        query['bool']['must'].push(old_queries);
        query['bool']['must'].push(new_queries);
        this._query = query;
        return this;
    }

    /**
     * Add a boolean must not query
     * @param conditions
     * @param should_or_must default is "must". Can be "must" or "should". Only needed when conditions is array
     * @returns {ElasticSearch}
     */
    mustNotQuery(conditions, should_or_must = "must") {
        let old_queries = this._query;
        this._query={};
        this.query(conditions, should_or_must);
        let new_queries = this._query;
        let query = null;
        if(old_queries.match_all === undefined) {
            query = {
                bool: {
                    must : [],
                    must_not: []
                },
            };
        } else {
            query = {
                bool: {
                    must_not: []
                },
            };
        }
        query['bool']['must_not'].push(new_queries);
        if(old_queries.match_all === undefined) {
            query['bool']['must'] = [old_queries];
        }
        this._query = query;
        return this;
    }

    /**
     * Maximum number of records to return
     * @method limit
     * @param limit
     * @return {ElasticSearch}
     */
    limit(limit) {
        this._limit = limit;
        return this;
    }

    /**
     * Number of records to skip
     * @method offset
     * @param offset
     * @return {ElasticSearch}
     */
    offset(offset) {
        this._offset = offset;
        return this;
    }

    /**
     * Search and return all records
     * @method all
     * @async
     * @return {Promise<Array>}
     */
    async all() {
        let params = buildQuery.call(this);
        let ret = [];
        let result = await this.constructor.client.search(params);
        if(result.hits) {
            let total = result.hits.total;
            let obj = [];
            if(total>0) {
                result.hits.hits.forEach(function(data){
                    let source = data._source;
                    obj.push(source);
                });
            }

            ret['total'] = total;
            ret['data'] = obj;
            return ret;
        } else {
            ret['total'] = 0;
            ret['data'] = [];
            return ret;
        }
    }

    /**
     * Completion suggester.
     * @static
     * @param keyword
     * @param field
     * @param regex
     * @param size
     * @param source properties to retrieve
     * @param options For available options, pls refer to https://www.elastic.co/guide/en/elasticsearch/reference/current/search-suggesters-completion.html
     * @return {*}
     */
    static async suggest(keyword, field, regex = false, source = null, size = 5, options = null) {
        let params = {
            index : this.index,
            type : this.type,
            body : {
                suggest: {}
            }
        };
        if(source !== null) {
            params['body']["_source"] = source;
        }
        params['body']['suggest'][this.type + "-suggest"] = {
            completion : {
                field : field,
                size : size
            }
        };
        if(regex === true) {
            params['body']['suggest'][this.type + "-suggest"]["regex"] = keyword;
        } else {
            params['body']['suggest'][this.type + "-suggest"]["prefix"] = keyword;
        }
        if(options!== null && typeof options === 'object' && Object.keys(options).length > 0) {
            Object.assign(params['body']['suggest'][this.type + "-suggest"]['completion'], options);
        }
        let result = await this.client.search(params);
        return result["suggest"][this.type + "-suggest"][0].options;
    }

    /**
     * Return first record of a given query
     * @method one
     * @async
     * @return {Promise<*>}
     */
    async one() {
        this._offset = 0;
        this._limit = 1;
        let params = buildQuery.call(this);
        let result = await this.constructor.client.search(params);
        if(result.hits) {
            let total = result.hits.total;
            let source = null;
            if(total>0) {
                source = result.hits.hits[0]['_source'];
            }
            return source;
        } else {
            return null;
        }
    }

    /**
     * Copy a JSON object to a new Model
     * @method _setAttributes
     * @static
     * @param data
     * @private
     */
    static _setAttributes(data) {
        let obj = new this();
        let mapping = this.mapping;
        let properties = mapping[this.type]['properties'];
        let primary_ids = [];
        for(var key in properties) {
            if(data[key]) {
                if(data[key] instanceof Date) {
                    obj[key] = data[key].getTime();
                } else {
                    obj[key] = data[key];
                }
                if(this.primaryKeys.includes(key)) {
                    primary_ids.push(data[key].toString());
                }
            } else if(this.primaryKeys.includes(key)) {
                throw new Error("Primay key " + key +" must have value");
            }
        }
        obj._primary_id = primary_ids.join("-");
        return obj;
    }

}

/**
 * @method getInsertParams
 * @return {{index: string, type: string, id: (id|string), body: Array}}
 */
function getInsertParams(refresh = false) {
    let params = {
        index: this.constructor.index,
        type: this.constructor.type,
        id: this._primary_id,
        refresh : refresh,
        body: {}
    };
    let mapping = this.constructor.mapping;
    let keys = Object.keys(mapping[this.constructor.type]['properties']);
    for(let key in this) {
        if(keys.indexOf(key) >= 0) {
            if(mapping[this.constructor.type]['properties'][key]['type'] == 'completion') {
                params['body'][key] = {'input' : this[key]};
            } else {
                params['body'][key] = this[key];
            }
        }
    }
    return params;
}

function getUpdateParams(refresh = false) {
    let params = {
        index: this.constructor.index,
        type: this.constructor.type,
        id: this._primary_id,
        refresh : refresh,
        body: {doc: {}},
    };
    let mapping = this.constructor.mapping;
    let keys = Object.keys(mapping[this.constructor.type]['properties']);
    for(let key in this) {
        let value = this[key];
        if(keys.indexOf(key) >= 0) {
            if (mapping[this.constructor.type]['properties'][key]["type"] == 'completion') {
                params['body']['doc'][key] = {'input': value};
            } else {
                params['body']['doc'][key] = value;
            }
        }
    }
    return params;
}

function resetFilters(should_or_must) {
    if(should_or_must === 'must') {
        this._filters = {
            bool : {
                must : []
            }
        };
    } else {
        this._filters = {
            bool : {
                should : []
            }
        };
    }
}

function add(refresh = false) {
    let params = getInsertParams.call(this, refresh);
    return this.constructor.client.index(params);
}

function update(refresh = false) {
    let params = getUpdateParams.call(this, refresh);
    return this.constructor.client.update(params);
}

function buildQuery() {
    let query = {};
    if(Object.keys(this._filters).length>0) {
        query = {
            bool: {
                filter: this._filters,
                must: this._query,
            }
        };
    } else {
        query = this._query;
    }
    let params = {
        index: this.constructor.index,
        type: this.constructor.type,
        body: {
            from: this._offset,
            size: this._limit,
            query: query,
            sort: this._sort,
        },
    };
    return params;
}

module.exports = ElasticSearch;